/*
 * Copyright (c) 1991-1998 Open Software Foundation, Inc. 
 *  
 * 
 */
/*
 * MkLinux
 */

#ifndef _ASM_OSFMACH3_PPC_BITOPS_H
#define _ASM_OSFMACH3_PPC_BITOPS_H

#include <asm-ppc/bitops.h>

#if 0
extern int find_first_zero_bit(void *addr, int size);
extern int find_next_zero_bit(void *addr, int len, int nr);
extern unsigned long ffz(unsigned long word);
#endif

#endif	/* _ASM_OSFMACH3_PPC_BITOPS_H */
