/*
 * Copyright (c) 1991-1998 Open Software Foundation, Inc. 
 *  
 * 
 */
/*
 * MkLinux
 */


#ifndef _ASM_OSFMACH3_MACHINE_BUGS_H
#define _ASM_OSFMACH3_MACHINE_BUGS_H

static void check_bugs(void)
{
}

#endif	/* _ASM_OSFMACH3_MACHINE_BUGS_H */


