/* $Id: isdn_cards.h,v 1.3 1998/11/18 23:32:36 trini Exp $

 * Linux ISDN subsystem, initialization for non-modularized drivers.
 *
 * Copyright 1994-1998 by Fritz Elfert (fritz@isdn4linux.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: isdn_cards.h,v $
 * Revision 1.3  1998/11/18 23:32:36  trini
 * Merge to 2.0.36
 *
 * Revision 1.2.2.1  1998/11/05 22:11:42  fritz
 * Changed mail-address.
 *
 * Revision 1.2  1997/02/03 23:31:55  fritz
 * Reformatted according CodingStyle
 *
 * Revision 1.1  1996/04/20 16:04:03  fritz
 * Initial revision
 *
 */

extern void isdn_cards_init(void);
