/* $Id: string.h,v 1.1.1.1 1998/10/02 21:17:13 root Exp $
 * string.h: External definitions for optimized assembly string
 *           routines for the Linux Kernel.
 *
 * Copyright (C) 1995 David S. Miller (davem@caip.rutgers.edu)
 */

#ifndef __SPARC_STRING_H__
#define __SPARC_STRING_H__

#define __HAVE_ARCH_BCOPY
#define __HAVE_ARCH_MEMMOVE
#define __HAVE_ARCH_MEMCPY
#define __HAVE_ARCH_MEMSET
#define __HAVE_ARCH_STRLEN

#endif /* !(__SPARC_STRING_H__) */
