/*
 * Copyright (c) 1991-1998 Open Software Foundation, Inc. 
 *  
 * 
 */
/*
 * MkLinux
 */

#ifndef _ASM_OSFMACH3_PPC_BUGS_H
#define _ASM_OSFMACH3_PPC_BUGS_H

#include <asm-ppc/bugs.h>

#endif	/* _ASM_OSFMACH3_PPC_BUGS_H */
