/*
 * Copyright (c) 1991-1998 Open Software Foundation, Inc. 
 *  
 * 
 */
/*
 * MkLinux
 */


#ifndef _ASM_OSFMACH3_MACHINE_MMU_CONTEXT_H
#define _ASM_OSFMACH3_MACHINE_MMU_CONTEXT_H

#endif	/* _ASM_OSFMACH3_MACHINE_MMU_CONTEXT_H */
