/*
 * Copyright (c) 1991-1998 Open Software Foundation, Inc.
 *
 *
 */
/*
 * MkLinux
 */

#ifndef _ASM_OSFMACH3_PPC_ADB_MOUSE_H
#define _ASM_OSFMACH3_PPC_ADB_MOUSE_H

#include <asm-ppc/adb_mouse.h>

#endif /* _ASM_OSFMACH3_PPC_ADB_MOUSE_H */
