/*
 * Copyright (c) 1991-1998 Open Software Foundation, Inc. 
 *  
 * 
 */
/*
 * MkLinux
 */

#ifndef	_OSFMACH3_MMU_CONTEXT_H_
#define _OSFMACH3_MMU_CONTEXT_H_

/*
 * get a new mmu context.. we don't know about contexts.
 */
#define get_mmu_context(x)	do { } while (0)

#endif	/* _OSFMACH3_MMU_CONTEXT_H_ */
