/*
 * Copyright (c) 1991-1998 Open Software Foundation, Inc. 
 *  
 * 
 */
/*
 * MkLinux
 */

#ifndef _ASM_OSFMACH3_I386_DMA_H
#define _ASM_OSFMACH3_I386_DMA_H

#include <asm/io.h>	/* need byte IO */

#endif	/* _ASM_OSFMACH3_I386_DMA_H */
