/*	Generated by configure. Don't edit!!!!	*/
/*	Making changes to this file is not as simple as it may look.	*/

/*	If you change the CONFIG_ settings in local.h you	*/
/*	_have_ to edit .defines too.	*/

#undef  CONFIG_UNUSED1
#undef  CONFIG_UNUSED2
#define CONFIG_AUDIO
#undef  CONFIG_MIDI

#undef  CONFIG_MPU_EMU
#define CONFIG_SEQUENCER

#define DSP_BUFFSIZE (16*1024)
#define SELECTED_SOUND_OPTIONS	0x00880002
#define SOUND_CONFIG_DATE "Fri Mar 7 17:37:08 PST 1997"
#define SOUND_CONFIG_BY "russb2"
#define SOUND_UNAME_A "Linux bergru2.apple.com 2.0.28-osfmach3 #1 Thu Feb 13 12:26:38 PST 1997 ppc"
