/*
 * Copyright (c) 1991-1998 Open Software Foundation, Inc. 
 *  
 * 
 */
/*
 * MkLinux
 */

#ifndef _ASM_OSFMACH3_MACHINE_PAGE_H
#define _ASM_OSFMACH3_MACHINE_PAGE_H

#define PAGE_SHIFT	12

#include <osfmach3/page.h>

#endif	/* _ASM_OSFMACH3_MACHINE_PAGE_H */
